#include "tm1637.h"

#include <array>

TM1637::TM1637(uint8_t dio, uint8_t clk, PIO pio) : m_pio(pio)
{
    gpio_pull_up(dio);
    gpio_pull_up(clk);
    pio_gpio_init(pio, dio);
    pio_gpio_init(pio, clk);
    init(dio, clk);
}

void TM1637::init(uint8_t dio, uint8_t clk)
{
    static constexpr size_t PULL_THRESHOLD = 32;

    m_stateMachine = static_cast<uint8_t>(pio_claim_unused_sm(m_pio, true));
    auto offset = pio_add_program(m_pio, &tm1637_program);
    m_stateMachineConfig = tm1637_program_get_default_config(offset);

    sm_config_set_sideset_pins(&m_stateMachineConfig, clk);
    uint32_t both_pins = (1 << clk) | (1 << dio);
    pio_sm_set_pins_with_mask(m_pio, m_stateMachine, both_pins, both_pins);
    pio_sm_set_pindirs_with_mask(m_pio, m_stateMachine, both_pins, both_pins);

    sm_config_set_out_pins(&m_stateMachineConfig, dio, 1);
    sm_config_set_set_pins(&m_stateMachineConfig, dio, 1);

    sm_config_set_out_shift(&m_stateMachineConfig, true, false, PULL_THRESHOLD);

    setClockDivider();

    pio_sm_init(m_pio, m_stateMachine, offset, &m_stateMachineConfig);
    pio_sm_set_enabled(m_pio, m_stateMachine, true);
}

void TM1637::setClockDivider()
{
    static constexpr size_t FREQUENCY = 45'000;
    static constexpr size_t MAX_DIVIDER_VALUE = 65'536;

    uint32_t system_frequency = clock_get_hz(clk_sys);
    float divider = static_cast<float>(system_frequency) / FREQUENCY;

    if (divider > MAX_DIVIDER_VALUE) {
        divider = MAX_DIVIDER_VALUE;
    } else if (divider < 1) {
        divider = 1;
    }

    sm_config_set_clkdiv(&m_stateMachineConfig, divider);
}

void TM1637::setColon([[maybe_unused]] bool on)
{
    static constexpr size_t CONTROL_DISPLAY = 0x80'00;

    m_isColon = on;

    if (on)
        m_currentSegments |= CONTROL_DISPLAY;
    else
        m_currentSegments &= ~CONTROL_DISPLAY;

    send4Bytes(m_currentSegments);
}

void TM1637::send4Bytes(uint32_t data)
{
    static constexpr size_t BIT_MASK = 0xFF'FF;
    static constexpr size_t SHIFT_POSITIONS = 16;

    static uint32_t data_1;
    static uint32_t data_2;

    data_1 = data & BIT_MASK;
    data_2 = data >> SHIFT_POSITIONS;

    pio_sm_put_blocking(m_pio, m_stateMachine, (data_1 << (2 * BYTE_SIZE)) + (WRITE_ADDRESS << BYTE_SIZE) + WRITE_MODE);
    pio_sm_put_blocking(m_pio, m_stateMachine, data_2 << (2 * BYTE_SIZE));

    pio_sm_put_blocking(m_pio, m_stateMachine, BRIGHTNESS_BASE + m_brightness);
}

void TM1637::displayIambicMode(Mode iambic)
{
    static constexpr size_t LEFT_BYTE_MASK = 0xFF'FF'00'00;

    uint16_t segments {0};

    switch (iambic) {
    case Mode::IambicA:
        segments = 0b0000'0000'0111'0111;
        break;
    case Mode::IambicB:
        segments = 0b0000'0000'0111'1100;
        break;
    case Mode::Straight:
        segments = 0b0000'0000'0110'1101;
        break;
    default:
        segments = 0b0000'0000'0000'0000;
        break;
    }

    m_currentSegments = (m_currentSegments & LEFT_BYTE_MASK) + segments;
    if (m_isColon) {
        setColon(m_isColon);
    }
    send4Bytes(m_currentSegments);
}

void TM1637::displaySpeed(uint8_t speed)
{
    static constexpr std::array<uint8_t, 16> DIGIT_TO_SEGMENTS {
        0x3F /* 0 */, 0x06 /* 1 */, 0x5B /* 2 */, 0x4F /* 3 */, 0x66 /* 4 */, 0x6D /* 5 */, 0x7D /* 6 */, 0x07 /* 7 */,
        0x7F /* 8 */, 0x6F /* 9 */, 0x77 /* A */, 0x7C /* B */, 0x39 /* C */, 0x5E /* D */, 0x79 /* E */, 0x71 /* F */};

    static constexpr size_t RIGHT_BYTE_MASK = 0x00'00'FF'FF;

    uint8_t speed_orig = speed;
    auto base = 10;
    auto max_one_digit = base - 1;
    uint32_t segments {0};
    uint32_t temp_segments {0};

    if (speed <= max_one_digit) {
        segments = DIGIT_TO_SEGMENTS.at(speed);
    }

    else {
        while (speed != 0) {
            temp_segments = DIGIT_TO_SEGMENTS.at(speed % base);
            speed /= base;
            segments = temp_segments + (segments << BYTE_SIZE);
        }
    }

    auto num_div = speed_orig / base;
    if (num_div == 0) {
        segments = segments << BYTE_SIZE;
    }

    m_currentSegments = (m_currentSegments & RIGHT_BYTE_MASK) + (segments << (2 * BYTE_SIZE));
    if (m_isColon) {
        setColon(m_isColon);
    }
    send4Bytes(m_currentSegments);
}
